from datetime import datetime
from functools import partial

import lmfit
import numpy as np
import panel as pn
import bokeh.models as bm
import bokeh.events as be
from bokeh.plotting import figure

from .parameter import InteractiveParameters
from .models import InteractiveModelMixin


class InteractiveGuessSession:
    def get_selected_component(self):
        """ Returns the Model object associated with the selected component """
        desired_component_label = self._selector_pane.value
        for component in self._model.components:
            # just return the first one...there shouldn't be duplicates, in theory
            if component.prefix + component.name == desired_component_label:
                return component
        else:
            return None

    def redraw(self):
        """ Redraw all lines on the plot """
        # appropriately update the components
        self._fig_change(self._redraw_components)

        # update the polygon selector
        self.draw_selector_for_component(None, None)

        # this is a hack to make the GUI values that are set by expressions update
        # in order to make this nice, the `value` machinery of `lmfit.Parameter` needs to be rewritten
        # with `param` and `param.depends`, so in the mean time, this is good enough.
        for p in self._params.values():
            if p.expr is not None and p.expr != "":
                p.value = p.value

    def _redraw_components(self):
        """ Redraw all components of this model and the model itself """
        for c, v in self._model.eval_components(params=self._params, x=self.plot_xs).items():
            self._comp_dss[c].data = {'x': self.plot_xs, 'y': v}

        ymodel = self._model.eval(params=self._params, x=self._x)
        self._model_ds.data = {'x': self._x, 'y': ymodel}
        self._res_ds.data = {'x': self._x, 'y': ymodel - self._data}

    def _fig_change(self, f):
        """ Safely call a function f that updates the Bokeh figure state """
        # if the figure isn't yet displayed, `document` is None and we just call the function
        # but if the figure is being displayed, we have to schedule the function execution with the server's IOLoop
        if self._fig.document is None:
            f()
        else:
            self._fig.document.add_next_tick_callback(f)

    def _set_selector_options(self):
        """ Set the options for the component selector """
        self._selector_pane.options = [c.prefix + c.name for c in self._model.components
                                       if isinstance(c, InteractiveModelMixin)] + ['None']

    def set_model(self, model, params=None):
        """ Sets the Model to be fit

        If the Model or its components are not interactive, this interface will still work, but the non-interactive
        components cannot be edited dragging the plot.
        Args:
            model: an interactive Model or CompositeModel instance
            params: a Parameters object to be used as an initial guess for the model
                if None, will use `model.guess` as default setting
        """

        self._model = model
        self._set_selector_options()
        self._set_params(params)

        # hide all existing data sources; Bokeh doesn't allow us to remove them...
        # this is probably a lag issue if you change the model too many times
        def hide_ds(rds):
            rds.data = {'x': [], 'y': []}

        for c, remove_ds in self._comp_dss.items():
            self._fig_change(partial(hide_ds, remove_ds))

        # draw all the components and keep all data sources & renderers
        self._comp_dss = {}
        self._comp_grs = {}

        def render_comp(rds, rc):
            self._comp_grs[rc] = self._fig.line('x', 'y', source=rds, line_dash='dashed')

        for c, v in model.eval_components(params=self._params, x=self.plot_xs).items():
            assert self.plot_xs.shape == getattr(v, 'shape', None), f"Given `self.plot_xs`, model component '{c}' was "\
                                                                    f"expected to evaluate to a numpy array of shape "\
                                                                    f"{self.plot_xs.shape} but it did not; cannot plot"
            ds = bm.ColumnDataSource({'x': self.plot_xs, 'y': v})
            self._fig_change(partial(render_comp, ds, c))
            self._comp_dss[c] = ds

        # we really only need to update the model, but this is fine
        self.redraw()

    def set_params(self, params):
        self._set_params(params)
        if self._params is not None:
            self.redraw()

    def _set_params(self, params):
        """ Private `set_params` that does not redraw """
        if params is None:
            if hasattr(self._model, 'guess') and callable(self._model.guess):
                self._params = self._model.guess(self._data, x=self._x)
            else:
                self._params = None
        else:
            self._params = params

        if self._params is not None:
            self._params.__class__ = InteractiveParameters
            self._params.on_noninteractive_update = lambda *args: self.set_params(self._params)
            self._param_pane = self._params.controls(value_callback=lambda *args: self.redraw())
            self._rhs_pane[-1] = self._param_pane

            # Clear the selector's default value
            self._selector_pane.value = 'None'

    def set_data(self, data, x=None):
        """ Sets the data to fit to the Model

        Args:
             data: the dependent variable data to fit to the model, as a 1D array
             x: the independent variable data to use in the fit.
                if None, will number the data points sequentially
        """
        self._set_data(data, x=x)
        self.redraw()

    def _set_data(self, data, x=None):
        """ Private `set_data` that does not redraw after update, for use in __init__ """
        x = np.arange(data.shape[0]) if x is None else x
        self._data = data
        self._x = x

        def update_data(): self._data_ds.data = {'x': self._x, 'y': self._data}

        self._fig_change(update_data)

    @classmethod
    def from_modelresult(cls, mr: lmfit.model.ModelResult, independent_var='x', **kwargs):
        """ Factory constructor to create an InteractiveGuessSession from a ModelResult

        If the `ModelResult` has been serialized, you may not be able to drag the models on the graph.
        This is because `ModelResult` will serialize its model as a generic Models, not the Interactive components.

        Args:
            mr: the `ModelResult` to open interactively
            independent_var: the name of the independent variable to be used for plotting.
                if None or not found, will number the data sequentially
            **kwargs: override any parameters you wish to be passed to `__init__`
        """
        x = mr.userkws.get(independent_var, None) if independent_var is not None else None
        args = dict(model=mr.model, data=mr.data, params=mr.params, x=x)
        args.update(kwargs)
        return cls(**args)

    def __init__(self, model, data, params=None, x=None, plot_xs=None):
        """ Allows 1D models to be fit interactively

        The update architecture is this:

            * When the parameters are updated, the plot redraws
            * When the table updates, the parameters are updated
            * When the plot is dragged, the parameters are updated
            * When the mouse is scrolled, the parameters are updated
        Args:
             model: the Model instance to fit
             data: the data to fit the model to
        Keyword Args:
             params: the initial parameters to draw the Model with.  If None, use parameters given by `model.guess`
             x: the x-axis data points for `data`.  If None, will number the data points sequentially
             plot_xs: the x-axis data points to evaluate the model and its components at.  If None, will use `x`
        """
        self._last_fit_result = None
        self._last_fit_result_time = None

        self._last_selector_pane_value = "None"

        # the bokeh figure
        self._fig = fig = figure()

        # set up the figure for the residuals
        self._resfig = figure()
        self._res_ds = bm.ColumnDataSource({'x': [], 'y': []})
        self._resfig.circle_dot('x', 'y', source=self._res_ds, fill_color='red', line_color=None)

        # --- Compose the main Panel application ---
        self._panel = pn.Row()
        _bokeh_pane = pn.pane.Bokeh(fig)
        self._panel.append(pn.Tabs(('Main Plot', pn.Column(_bokeh_pane,
                                                           pn.pane.Markdown("### Residuals"),
                                                           self._resfig)),
                                   ('Residuals Only', self._resfig)))
        self._tabs_pane = pn.Tabs()
        self._panel.append(self._tabs_pane)
        # add the component-edit selector, along with a placeholder for the param editor table
        self._selector_pane = pn.widgets.Select(name="Editing Component", options=['None'])
        self._rhs_pane = pn.Column(self._selector_pane, pn.pane.Str("param pane placeholder"))
        self._tabs_pane.append(("Edit", self._rhs_pane))
        # add the fit button
        self._fit_pane = pn.widgets.Button(name="Fit")
        self._fit_results_pane = pn.pane.Str()
        self._tabs_pane.append(("Fit", pn.Column(self._fit_pane, self._fit_results_pane)))

        # --- Setup & draw functional internal components ---
        # setup the polygon selector tools & renderers for drawing
        self._pdt_ds = bm.ColumnDataSource({'x': [], 'y': []})
        pdt_renderer = fig.multi_line('x', 'y', source=self._pdt_ds, line_color='green')
        pdt = bm.PolyEditTool(renderers=[pdt_renderer], vertex_renderer=fig.circle([], [], size=10, color='green'))
        fig.add_tools(pdt)
        # setup raw data & the independent variable x
        self._data_ds = bm.ColumnDataSource({'x': [], 'y': []})
        self._x = None
        self._data = None
        self._set_data(data, x=x)  # will appropriately init above 3 variables
        self.plot_xs = self._x if plot_xs is None else plot_xs  # depends on _x
        fig.circle_dot('x', 'y', source=self._data_ds, fill_color='red', line_color=None)
        # setup the model, params, and associated model parameter table
        self._model = None
        self._params = None
        self._param_pane = None
        self._model_ds = bm.ColumnDataSource({'x': [], 'y': []})
        self._comp_dss = {}  # component data sources
        self._comp_grs = {}  # component glyph renderers
        self.set_model(model, params=params)  # will appropriately init above 6 variables
        self._set_selector_options()  # now that we have the model, update the component selector
        fig.line('x', 'y', source=self._model_ds)

        # --- Setup interactive callbacks ---
        # callback to draw the polygon selector when a component is selected to edit
        self._selector_pane.link(_bokeh_pane, callbacks={'value': self.draw_selector_for_component})
        # callback to redraw the component when the polygon selector changes
        self._pdt_ds.on_change('data', self.update_component_from_selector)
        # callback to redraw the component & polygon selector when there is a scroll event
        fig.on_event(be.MouseWheel, self.update_component_params_from_scroll)
        # callback to run the fit when clicked
        self._fit_pane.on_click(self.update_component_params_from_fit)
        # callback to set the selector to None when we click away from the edit tab
        self._tabs_pane.param.watch(self.update_selector_pane_from_tab_select, ['active'])
        # callback to disable window scrolling when we are editing a component
        # noinspection PyTypeChecker
        self._selector_pane.jscallback(args={'selector': self._selector_pane}, value="""
        // need to hold on to function reference, so we can't redefine it
        if (!selector.hasOwnProperty('__hopefully_disused_name'))
           selector.__hopefully_disused_name = function(e) { e.preventDefault(); }
        
        if (selector.value == "None") {
           window.removeEventListener('wheel', selector.__hopefully_disused_name, { passive: false });
        } else {
           window.addEventListener('wheel', selector.__hopefully_disused_name, { passive: false });
        }
        """)  # alternatively, this could be done when the mouse enters/leaves the Bokeh plot area...
        # also not really sure why this doesn't suppress events to Bokeh, but it doesn't, which is nice

        self.thread = None

    def update_component_params_from_fit(self, event):
        """ Callback to fit & update our parameters due to a fit request """
        self._fit_pane.button_type = 'warning'
        self._fit_pane.name = "fit in progress..."

        result = self._model.fit(self._data, x=self._x, params=self._params)
        self._last_fit_result = result
        self._last_fit_result_time = datetime.now()

        # update fit results pane
        timeline = f"Fit Results from {self._last_fit_result_time}\n"
        self._fit_results_pane.object = timeline + self._last_fit_result.fit_report()

        self._params.update_existing_values(result.params)
        self._fit_pane.button_type = 'default'
        self._fit_pane.name = "Fit"

    def update_component_params_from_scroll(self, event):
        """ Callback to update the component when the scroll wheel is spun """
        component = self.get_selected_component()
        if component is None:
            return
        new_params = component.update_params_from_scroll(event.delta, params=self._params)
        self._params.update_existing_values(new_params)
        # redraw is triggered by the parameters updating

    def update_component_from_selector(self, attr, old, new):
        """ Callback to update the component when the polygon selector is dragged"""
        if attr != 'data':
            return
        points = dict(self._pdt_ds.data)
        component = self.get_selected_component()
        if component is None:
            return

        new_params = component.from_poly_selector_shape(points['x'][0], points['y'][0])
        # plain update will overwrite the Parameter objects, destroying our links
        self._params.update_existing_values(new_params)

    def draw_selector_for_component(self, target, event):
        """ Callback to redraw the polygon selector when a component is selected for editing """

        def subfunc():
            component = self.get_selected_component()
            if component is None:
                # we didn't find the component (so it's probably None), so don't show any points
                self._pdt_ds.data = dict(x=[[]], y=[[]])
            else:
                x, y = component.to_poly_selector_shape((self._x.min(), self._x.max()), self._params)
                self._pdt_ds.data = dict(x=[x], y=[y])

            for c, gr in self._comp_grs.items():
                gr.glyph.line_color = 'green' if c == getattr(component, 'prefix', None) else '#1F77B4'

        # since we screw with the figure directly, this needs to be wrapped
        self._fig_change(subfunc)

    def update_selector_pane_from_tab_select(self, *events):
        """ Callback to manage disabling the editing features when we switch out of the edit tab """
        # this is mostly just so the scroll wheel works when you switch tabs
        if self._tabs_pane.active == 0:
            # restores previous value
            self._selector_pane.value = self._last_selector_pane_value
        else:
            self._last_selector_pane_value = self._selector_pane.value
            self._selector_pane.value = 'None'

    def do_fit(self):
        """ Perform a fit of the current Model with current parameters

        Returns:
            a `ModelResult` object describing the fit
        """
        self.update_component_params_from_fit(None)
        return self._last_fit_result

    @property
    def panel(self):
        """ The main Panel application associated with this guess session """
        return self._panel

    @property
    def params(self):
        """ The current parameters that have been chosen """
        return self._params

    @property
    def model(self):
        """ The model current being fit """
        return self._model

    @property
    def data(self):
        """ The dependent data (y-axis) currently being fit """
        return self._data

    @property
    def residuals(self):
        """ The residuals of the current model & data being fit """
        return self._model.eval(params=self._params, x=self._x) - self._data

    @property
    def x(self):
        """ The independent data (x-axis) currently being fit """
        return self._x

    @property
    def fit(self):
        """ The `ModelResult` object from the last fit performed """
        return self._last_fit_result

    def start(self, threaded=True, **kwargs):
        """ Starts the front-end Panel application on a separate thread """
        self.thread = self._panel.show(threaded=threaded, **kwargs)

    def stop(self):
        """ Stops the front-end Panel application """
        if self.thread is not None:
            self.thread.stop()

    def is_alive(self):
        """ Returns whether the front-end Panel application is still running """
        if self.thread is not None:
            return self.thread.is_alive()

    def join(self):
        """ Blocks until the front-end application shuts down """
        if self.thread is not None:
            self.thread.join()



import os
import lmfit


def save(obj, fname, overwrite='ask', **kwargs):
    """ Saves a `dump`-able object to a file.

    This is designed to reduce boilerplate at the command line when saving objects like
    `ModelResult`, `Model`, and `Parameters`
    """
    assert hasattr(obj, 'dump') and callable(obj.dump), "Object must have a dump method"
    mode = 'x'
    if os.path.exists(fname):
        if overwrite == 'ask' and input("Overwrite? (y/n)") == 'y':
            mode = 'w'
        elif overwrite is True:
            mode = 'w'
    else:
        mode = 'w'

    with open(fname, mode=mode) as f:
        obj.dump(f, **kwargs)


def load(fname, kind, **kwargs):
    """ Loads an lmfit object from disk

    This is designed to reduce boilerplate at the command line.
    Args:
         fname: the file name to open
         kind: the kind of object that is being opened: 'model', 'modelresult', or 'parameters'
         **kwargs: passed to the underlying load functions
    """
    kind = str(kind).lower().strip()
    if kind == 'model':
        return lmfit.model.load_model(fname, **kwargs)
    elif kind == 'modelresult':
        return lmfit.model.load_modelresult(fname, **kwargs)
    elif kind == 'parameters':
        with open(fname, 'r') as f:
            return lmfit.Parameters().load(f, **kwargs)
    else:
        raise ValueError(f"Unrecognized kind {kind}; try model, modelresult, or parameters")



from abc import ABC, abstractmethod
from typing import Tuple
import operator

import lmfit
import numpy as np


class InteractiveModelMixin(lmfit.models.Model, ABC):
    def interactive_guess(self, data, x=None, params=None, plot_xs=None):
        """ Like `guess`, but with an attached interactive front-end """
        from .interactive_guess import InteractiveGuessSession
        igs = InteractiveGuessSession(self, data, x=x, params=params, plot_xs=plot_xs)
        igs.start()
        return igs

    def update_params_from_scroll(self, ticks, params=None, **kwargs):
        return self._update_params_from_scroll(ticks, self.make_funcargs(params, **kwargs))

    def _update_params_from_scroll(self, ticks, vals):
        """ Describes how parameters should be updated from a scroll event

        Should be implemented by interactive component subclasses
        Args:
            ticks: the number of scroll wheel ticks on this event
            vals: a dictionary describing the current function parameter values for this model
        Returns:
            A Parameters object describing this model for the given selector
        """
        raise NotImplementedError

    def to_poly_selector_shape(self, x_range, params=None, **kwargs) -> Tuple[list, list]:
        return self._to_poly_selector_shape(x_range, self.make_funcargs(params, **kwargs))

    def _to_poly_selector_shape(self, x_range, vals) -> Tuple[list, list]:
        """ Transform given model parameters into a poly selector

        Should be implemented by interactive component subclasses
        Args:
            x_range: a 2-tuple with the min and max values for the independent coordinate
            vals: a dictionary describing the current function parameter values for this model
        Returns:
            Two lists of the same length, describing the x-points and the y-points to draw the polygon at
        """
        raise NotImplementedError

    def from_poly_selector_shape(self, xs, ys):
        return self._from_poly_selector_shape(xs, ys)

    def _from_poly_selector_shape(self, xs, ys):
        """ Transforms a poly selector to parameters for this model

        Should be implemented by interactive component subclasses
        Args:
            xs: the x-coordinate data for a polygon selector
            ys: the y-coordinate data for a polygon selector
        Returns:
            A Parameters object describing this model for the given selector
        """
        raise NotImplementedError

    def __add__(self, other):
        """+"""
        return ICompositeModel(self, other, operator.add)

    def __sub__(self, other):
        """-"""
        return ICompositeModel(self, other, operator.sub)

    def __mul__(self, other):
        """*"""
        return ICompositeModel(self, other, operator.mul)

    def __div__(self, other):
        """/"""
        return ICompositeModel(self, other, operator.truediv)

    def __truediv__(self, other):
        """/"""
        return ICompositeModel(self, other, operator.truediv)


# example of what a model component needs to implement to be interactive
class ILinearModel(InteractiveModelMixin, lmfit.models.LinearModel):
    def _update_params_from_scroll(self, ticks, vals):
        vals['slope'] += ticks/1400
        return self.make_params(**vals)

    def _to_poly_selector_shape(self, x_range, vals):
        return list(x_range), list(self.eval(x=np.array(x_range), **vals))

    def _from_poly_selector_shape(self, xs, ys):
        m = (ys[1] - ys[0])/(xs[1] - xs[0])
        b = ys[0] - (xs[0] * m)
        return self.make_params(slope=m, intercept=b)


class InteractiveDistributionMixin(InteractiveModelMixin):
    def _update_params_from_scroll(self, ticks, vals):
        scale = np.clip((ticks / 700) + 1, 0.5, 2)
        vals['sigma'] *= scale
        vals['amplitude'] *= scale  # don't change peak value
        return self.make_params(**vals)

    def _to_poly_selector_shape(self, x_range, vals):
        c = vals['center']
        s = vals['sigma']
        xs = np.array([c - s, c])  # pick the peak & one sigma to the left of the peak
        ys = self.eval(x=xs, **vals)
        return list(xs), list(ys)

    def _from_poly_selector_shape(self, xs, ys):
        center = xs[-1]
        sigma = abs(xs[-1] - xs[0])
        amplitude = ys[-1] * sigma / self.height_factor
        return self.make_params(amplitude=amplitude, center=center, sigma=sigma)


class IGaussianModel(InteractiveDistributionMixin, lmfit.models.GaussianModel):
    pass


class ILorentzianModel(InteractiveDistributionMixin, lmfit.models.LorentzianModel):
    pass


class IConstantModel(InteractiveModelMixin, lmfit.models.Model):
    """ A ConstantModel that behaves as expected with numpy arrays; see lmfit#684 for details """
    def __init__(self, independent_vars=['x'], prefix='', nan_policy='raise',
                 **kwargs):
        kwargs.update({'prefix': prefix, 'nan_policy': nan_policy,
                       'independent_vars': independent_vars})

        def constant(x, c=0.0):
            return c * np.ones(x.shape)
        super().__init__(constant, **kwargs)

    def guess(self, data, **kwargs):
        pars = self.make_params()
        pars['%sc' % self.prefix].set(value=data.mean())
        return lmfit.models.update_param_vals(pars, self.prefix, **kwargs)

    def _update_params_from_scroll(self, ticks, vals):
        vals['c'] += ticks/700
        return self.make_params(**vals)

    def _to_poly_selector_shape(self, x_range, vals):
        return [sum(x_range)/2], [vals['c']]

    def _from_poly_selector_shape(self, xs, ys):
        return self.make_params(c=ys[0])


# composite model isn't a component, so no need to implement any methods.  this is just a preferred change
class ICompositeModel(InteractiveModelMixin, lmfit.model.CompositeModel):
    def guess(self, data, **kws):
        p = lmfit.Parameters()
        p.update(self.left.guess(data, **kws))
        p.update(self.right.guess(data, **kws))
        return p



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ivfit']

package_data = \
{'': ['*']}

install_requires = \
['bokeh>=2.3.1,<3.0.0', 'lmfit>=1.0.2,<2.0.0', 'panel>=0.11.3,<0.12.0']

setup_kwargs = {
    'name': 'ivfit',
    'version': '0.1.0',
    'description': 'Interactive, visual, fitting toolkit',
    'long_description': None,
    'author': 'Jonathan Okasinski',
    'author_email': 'jonathan.okasinski@gmail.com',
    'maintainer': 'Jonathan Okasinski',
    'maintainer_email': 'jonathan.okasinski@gmail.com',
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
}


setup(**setup_kwargs)



from functools import partial

import lmfit
import param
import panel as pn
import numpy as np

param_fields = ['value', 'min', 'max', 'vary', 'expr', 'brute_step', 'user_data']
max_float = 1e308


def clip(n):
    return np.clip(n, -max_float, max_float)


class InteractiveParameter(param.Parameterized, lmfit.Parameter):
    name = param.String()
    min = param.Number(allow_None=True)
    max = param.Number(allow_None=True)
    vary = param.Boolean()
    _expr = param.String(allow_None=True)

    def __init__(self, *args, **kwargs):
        lmfit.Parameter.__init__(self, *args, **kwargs)
        super().__init__(name=self.name, min=self.min, max=self.max, vary=self.vary)

    @property
    def clipped_min(self):
        """ Tracks `min`, but is JSON-compliant

        This means that it is clipped to +- `max_float`.
        When it is set, values greater or equal to `max_float` are interpreted as infinities.
        If it is set with a NaN value, it will interpret it as -inf
        """
        return clip(self.min)

    @clipped_min.setter
    def clipped_min(self, clipped_min):
        if clipped_min is None:
            self.min = -np.inf
        elif clipped_min >= max_float:
            self.min = np.inf
        elif clipped_min <= -max_float:
            self.min = -np.inf
        else:
            self.min = clipped_min

    @property
    def clipped_max(self):
        """ Tracks `max`, but is JSON-compliant

        See `clipped_min` for details.
        If this is set to a NaN value, it will be interpreted as +inf
        """
        return clip(self.max)

    @clipped_max.setter
    def clipped_max(self, clipped_max):
        if clipped_max is None:  # if you type in an out of range value JS-side, you'll get None here...
            self.max = np.inf
        elif clipped_max >= max_float:
            self.max = np.inf
        elif clipped_max <= -max_float:
            self.max = -np.inf
        else:
            self.max = clipped_max

    @lmfit.Parameter.value.setter
    def value(self, value):
        # this prevents serialization errors
        if abs(value) > max_float:
            raise ValueError(f"Absolute value of {value} must be less than {max_float}")
        lmfit.Parameter.value.fset(self, value)
        # call our callback if assigned when we are set
        if hasattr(self, "on_value_change") and callable(self.on_value_change):
            self.on_value_change(value)

        # ostensibly this should be done with a Dynamic parameter, or an @depends watcher
        # there's an edge case if someone sets self._val, we won't catch that, but _getval() will of course change
        # however, currently, this only happens during __init__ so this callback works well enough
        # (yes, this is fragile wrt updates to lmfit)

    @classmethod
    def from_parameter(cls, parameter):
        """ Factory method to create InteractiveParameter from a Parameter

        Only the name and the fields defined in `param_fields` are preserved.
        """
        o = cls(parameter.name)
        for a in param_fields:
            setattr(o, a, getattr(parameter, a))
        return o


class InteractiveParameters(lmfit.Parameters):
    # noinspection PyTypeChecker
    def controls(self, value_callback=None):
        """ Produces a control Panel for these parameters """
        panel = pn.GridSpec()

        # make the header row
        cols = ["Name", "Value", "Min", "Max", "Vary", "Expr"]
        for n, col in enumerate(cols):
            panel[0, n] = pn.pane.Markdown(f"### {col}", height_policy='min', width_policy='min', width=50)

        num_kws = dict(height=30, height_policy='fit')

        def apply(p, **kws):
            for k, v in kws.items():
                setattr(p, k, v)
            for k in ['width', 'height', 'width_policy', 'height_policy']:
                p.param[k].readonly = True

        # set all non-interactive parameters to interactive ones
        dnc = getattr(self, '_disable_nis_callback', False)
        self._disable_nis_callback = True
        try:
            for k, v in self.items():
                if not isinstance(v, InteractiveParameter):
                    self[k] = InteractiveParameter.from_parameter(v)
        finally:
            self._disable_nis_callback = dnc

        # compose the entire widget grid going row by row.  Each row is one parameter
        for n, (name, lfparam) in enumerate(self.items()):
            n += 1  # row index offset for the header

            # probably should migrate this code to InteractiveParameter, but no pressing reason yet

            # Add the parameter name
            panel[n, 0] = pn.pane.Str(name, width=100, width_policy='fit')

            # Add the parameter value
            value_pane = pn.widgets.FloatInput(value=clip(lfparam.value), width=100, width_policy='min', format='0.00')

            value_pane.link(lfparam, value='value')  # when the GUI updates, set the parameter value
            def value_update(p, v):
                p.value = v

            lfparam.on_value_change = partial(value_update, value_pane)  # when the parameter updates, update the GUI
            if callable(value_callback):
                value_pane.link(lfparam, callbacks={'value': value_callback})  # external callback for any value updates
            panel[n, 1] = value_pane

            # Add the parameter min
            min_pane = pn.widgets.FloatInput(value=clip(lfparam.min), width=100, width_policy='min')
            # when the GUI updates, update the parameter minimum.  interpret out of range values as -inf
            min_pane.link(lfparam, value='clipped_min')
            # when the parameter minimum changes, update the GUI with an in-range value.
            # (so this could be called spuriously, for example if the min goes from -1e308 to -inf)
            def update(p, lfp, *events): p.value = lfp.clipped_min
            lfparam.param.watch(partial(update, min_pane, lfparam), ['min'])
            panel[n, 2] = min_pane

            # Add the parameter max identically to the minimum
            max_pane = pn.widgets.FloatInput(value=clip(lfparam.max), width=100, width_policy='min')
            max_pane.link(lfparam, value='clipped_max')
            def update(p, lfp, *events): p.value = lfp.clipped_max
            lfparam.param.watch(partial(update, max_pane, lfparam), ['max'])
            panel[n, 3] = max_pane

            # Add the parameter vary
            vary_pane = pn.widgets.Checkbox(value=lfparam.vary, width_policy='min', width=50)
            vary_pane.link(lfparam, value='vary', bidirectional=True)  # no funny business as we have two params
            panel[n, 4] = vary_pane

            # Add the parameter expr
            expr_pane = pn.widgets.TextInput(value=lfparam.expr, width=150, width_policy='min')
            expr_pane.link(lfparam, value='expr')  # when the GUI updates, update the parameter value
            def update(p, lfp, *events): p.value = lfp.expr
            # public expr isn't a param, so we watch the plain private expr.
            # however this should be robust, as we are still using values from the public value
            lfparam.param.watch(partial(update, expr_pane, lfparam), ['_expr'])
            panel[n, 5] = expr_pane

            # now do some misc callbacks
            # we don't have a proper watcher on value, but changing expr can change the calculated value
            # so update that
            def update(p, lfp, *events): p.value = lfp.value
            expr_pane.link(lfparam, callbacks={'value': partial(update, value_pane, lfparam)})

            # disable value pane when there's an expression
            def update(ep, panes, *events):
                disabled = not(ep.value == "" or ep.value is None)
                for p in panes:
                    p.disabled = disabled
            expr_pane.link(value_pane, callbacks={'value': partial(update, expr_pane,
                                                                   [value_pane, min_pane, max_pane, vary_pane])})
            # run these misc callbacks
            expr_pane.param.trigger('value')

        for pane in panel:
            apply(pane, **num_kws)
        return panel

    def update_existing_values(self, params):
        """ Updates the value of all existing Parameter objects from supplied `params`

        Does not change the internal Parameter objects
        Does not update the value if an expression is already set
        """
        for c, p in params.items():
            if c in self and self[c].expr is None:
                self[c].value = p.value

    def dumps(self, **kws):
        # we aren't ourselves serializable nicely (callbacks?), so we make a plain non-interactive copy to serialize
        return super().__class__.dumps(self.copy(), **kws)

    def __setitem__(self, key, new):
        if new is not None and not isinstance(new, lmfit.Parameter):
            raise ValueError("'%s' is not a Parameter" % new)

        # this is supposed to make parameters less fragile by making it harder to overwrite the objects directly
        # and calling a callback when we are updated with new, non-interactive parameters
        # I'm not confident that this covers all edge cases; experience would suggest it doesn't, but I can't name any
        # at the moment

        old = self.get(key, None)
        old_int = isinstance(old, InteractiveParameter)
        new_int = isinstance(new, InteractiveParameter)

        if old_int and (not new_int):
            # we are replacing an interactive parameter with a non-interactive one
            # if the names are the same then copy over the attributes
            # no callback needed as we avoid replacing the parameter at all
            if old.name == new.name:
                for f in param_fields:  # this is sort of fragile against new lmfit features...
                    setattr(self[key], f, getattr(new, f))
                return

        super().__setitem__(key, new)
        if hasattr(self, '_disable_nis_callback') and self._disable_nis_callback:
            return
        if hasattr(self, 'on_noninteractive_update') and callable(self.on_noninteractive_update):
            self.on_noninteractive_update()

    def add_many(self, *parlist):
        dnc = getattr(self, '_disable_nis_callback', False)
        self._disable_nis_callback = True
        super().add_many(*parlist)
        self._disable_nis_callback = dnc

        # for performance reasons, just call this once rather than after every __setitem__ as
        # GUI redrawing is pretty expensive.  there's no other reason for this method override
        if hasattr(self, 'on_noninteractive_update') and callable(self.on_noninteractive_update):
            self.on_noninteractive_update()
